package main

import (
	"bytes"
	"fmt"
	"html"
	"log"
	"net/http"
	"net/url"
	"strings"
)

func (this *Application) Templatepage(w http.ResponseWriter, r *http.Request, pageDesc, extraHead string, cb func()) {

	pageTitle := this.cfg.Template.AppName
	if pageDesc != "" {
		pageTitle = pageDesc + ` | ` + pageTitle
	}

	w.Header().Set(`Content-Type`, `text/html; charset=UTF-8`)
	w.WriteHeader(200)
	fmt.Fprint(w, `<!DOCTYPE html>
<html lang="en">
	<head>
		<meta http-equiv="X-UA-Compatible" content="IE=edge">
		<meta charset="UTF-8">
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
		<meta name="viewport" content="width=device-width, initial-scale=1">
		<title>`+html.EscapeString(pageTitle)+`</title>
		`+extraHead+`
		<link rel="shortcut icon" href="/static/logo.png" type="image/png">
		<link rel="apple-touch-icon" href="/static/logo.png" type="image/png">
		<link type="text/css" rel="stylesheet" href="/static/style.css">
	</head>
	<body>
		<div id="container">
			<div id="content">
				<h1><a href="/"><div id="logo"></div>`+html.EscapeString(this.cfg.Template.AppName)+`</a></h1>
	`)
	cb()
	fmt.Fprint(w, `
	</body>
	<script type="text/javascript" src="/static/site.js"></script>
</html>
`)

}

func (this *Application) internalError(w http.ResponseWriter, r *http.Request, err error) {
	log.Printf("%s %s: %s", r.Method, r.URL.Path, err)
	http.Error(w, "An internal error occurred.", 500)
}

func (this *Application) Delay(w http.ResponseWriter, r *http.Request) {
	this.Templatepage(w, r, "Loading...", "", func() {
		fmt.Fprintf(w, `
		<h2>Loading, please wait...</h2>
		
		<meta http-equiv="refresh" content="5">
		`)
	})
}

func (this *Application) Homepage(w http.ResponseWriter, r *http.Request) {

	this.reposMut.RLock()
	defer this.reposMut.RUnlock()

	if len(this.reposCache) == 0 {
		// We haven't loaded the repositories from Gitea yet
		this.Delay(w, r)
		return
	}

	// Ready for template

	this.Templatepage(w, r, "", "", func() {
		fmt.Fprint(w, `
		`+this.cfg.Template.HomepageHeaderHTML+`		
		<select id="sortorder" style="float:right;">
			<option value="data-sort-al">Alphabetical</option>
			<option value="data-sort-ls">Lifespan</option>
			<option value="data-sort-ct" selected>Recent projects</option>
			<option value="data-sort-mt">Recent updates</option>
		</select>
		
		<h2>Projects <small>(`+fmt.Sprintf("%d", len(this.reposCache))+`)</small></h2>
		<table id="projtable-main" class="projtable">
`)
		for _, repo := range this.reposCache {
			pageHref := html.EscapeString(`/` + url.PathEscape(repo.Name))

			normalisedDesc := repo.Description
			normalisedDesc = strings.TrimRight(repo.Description, `.`)
			if len(normalisedDesc) > 0 {
				// Lowercase the first letter of the description, unless it starts with an acronym (all letters uppercase first word) or CamelCase word
				firstWord := strings.SplitN(normalisedDesc, " ", 2)[0]
				isAcronymOrCamelCase := len(firstWord) > 1 && (firstWord[1:] != strings.ToLower(firstWord[1:]))

				if !(isAcronymOrCamelCase || firstWord == `Go`) {
					normalisedDesc = strings.ToLower(normalisedDesc[0:1]) + normalisedDesc[1:]
				}

				// Add leading `<COMMA><SPACE>` to separate from the repo title
				normalisedDesc = `, ` + normalisedDesc
			}

			rowClass := ""
			for _, topic := range repo.topics {
				rowClass += `taggedWith-` + topic + ` `
			}

			fmt.Fprint(w, `
			<tr
				class="`+html.EscapeString(rowClass)+`"
				data-sort-al="`+fmt.Sprintf("-%d", this.reposAlphabeticalOrder[repo.Name])+`"
				data-sort-ls="`+fmt.Sprintf("%.0f", repo.newestCommit.Sub(repo.GiteaCreated).Seconds())+`"
				data-sort-ct="`+fmt.Sprintf("%d", repo.GiteaCreated.Unix())+`"
				data-sort-mt="`+fmt.Sprintf("%d", repo.newestCommit.Unix())+`"
			>
				<td>
					<a href="`+pageHref+`"><img class="homeimage" loading="lazy" src="`+html.EscapeString(`/:banner/`+url.PathEscape(repo.Name))+`"></div></a>
				</td>
				<td>
					<strong>`+html.EscapeString(repo.Name)+`</strong>`+html.EscapeString(normalisedDesc)+`
					<a href="`+pageHref+`" class="article-read-more">more...</a>
					<br>
					<small>
			`)
			for _, topic := range repo.topics {
				fmt.Fprint(w, `<a class="tag tag-link" data-tag="`+html.EscapeString(topic)+`">`+html.EscapeString(topic)+`</a> `)
			}
			fmt.Fprint(w, `
					</small>
				</td>
			</tr>
			`)
		}
		fmt.Fprint(w, `
		</table>
		`)
	})
}

func (this *Application) Bannerpage(w http.ResponseWriter, r *http.Request, repoName string) {
	ctx := r.Context()

	images, err := this.imageFilesForRepo(ctx, repoName)
	if err != nil {
		this.internalError(w, r, fmt.Errorf("listing images: %w", err))
		return
	}

	if len(images) == 0 {
		w.Header().Set(`Location`, `/static/no_image.png`)
		w.WriteHeader(301)
		return
	}

	w.Header().Set(`Location`, images[0].RawURL)
	w.WriteHeader(301)
}

func (this *Application) Repopage(w http.ResponseWriter, r *http.Request, repoName string) {
	ctx := r.Context()
	repoURL := this.cfg.Gitea.URL + url.PathEscape(this.cfg.Gitea.Org) + `/` + url.PathEscape(repoName)
	extraHead := ""

	readme, err := this.repoFile(ctx, repoName, `README.md`)
	if err != nil {
		this.internalError(w, r, fmt.Errorf("loading README.md: %w", err))
		return
	}

	lines := strings.Split(string(readme), "\n")

	// Check if this repo has the 'article' tag
	hasArticleTag := false
	this.reposMut.RLock()
	for _, rr := range this.reposCache {
		if rr.Name != repoName {
			continue
		}
		for _, topic := range rr.topics {
			if topic == "article" {
				hasArticleTag = true
				break
			}
		}
	}
	this.reposMut.RUnlock()

	// We add some extra badges based on special text entries
	extraBadgesMd := ``
	if !hasArticleTag {
		extraBadgesMd += ` %%REPLACEME__BADGE%%`
	}
	extraBadgesMd += ` [![](https://img.shields.io/badge/vcs-git-green?logo=git)](` + repoURL + `)`

	// Inject more badges to 3rd line; or, create badges on 3rd line if there are none already
	if len(lines) >= 3 && strings.Contains(lines[2], `shields.io`) {
		lines[2] += ` ` + extraBadgesMd
	} else {
		// Push other lines down
		lines = append([]string{lines[0], lines[1], extraBadgesMd, ""}, lines[2:]...)
	}

	readmeHtml, err := this.renderMarkdown(ctx, repoName, strings.Join(lines, "\n"))
	if err != nil {
		this.internalError(w, r, fmt.Errorf("rendering markdown: %w", err))
		return
	}

	readmeHtml = []byte(strings.Replace(string(readmeHtml), `%%REPLACEME__BADGE%%`, `<img src="/static/build_success_brightgreen.svg" style="width:90px;height:20px;">`, 1))

	images, err := this.imageFilesForRepo(ctx, repoName)
	if err != nil {
		this.internalError(w, r, fmt.Errorf("listing images: %w", err))
		return
	}

	// If the Git repository contains a top-level go.mod file, allow vanity imports
	if goMod, err := this.repoFile(ctx, repoName, `go.mod`); err == nil {

		// Check the first line should be `module MODULENAME\n`
		firstLine := bytes.SplitN(goMod, []byte("\n"), 2)[0]
		if bytes.HasPrefix(firstLine, []byte("module ")) {
			moduleName := firstLine[7:]
			extraHead = `<meta name="go-import" content="` + html.EscapeString(string(moduleName)) + ` git ` + repoURL + `.git">`
		}
	}

	// De-escalate all headers in rendered markdown to match our style
	repl := strings.NewReplacer(`<h1`, `<h2`, `<h2`, `<h3`, `<h3`, `<h4`,
		`</h1>`, `</h2>`, `</h2>`, `</h3>`, `</h3>`, `</h4>`)

	// Ready for template

	this.Templatepage(w, r, repoName, extraHead, func() {

		projBodyclass := `projbody`
		if len(images) > 0 {
			projBodyclass += ` projbody_halfw`
		}

		fmt.Fprint(w, `<div class="projinfo"><div class="`+projBodyclass+`">`)
		repl.WriteString(w, string(readmeHtml))
		fmt.Fprint(w, `</div>`)

		if len(images) > 0 {
			fmt.Fprint(w, `<div class="projimg">`)
			for _, img := range images {
				fmt.Fprint(w, `<a href="`+html.EscapeString(img.RawURL)+`"><img alt="" class="thumbimage" src="`+html.EscapeString(img.RawURL)+`" /></a>`)
			}
			fmt.Fprint(w, `</div>`)
		}

		fmt.Fprint(w, `<div style="clear:both;"></div>`)
		fmt.Fprint(w, `</div>`) // projbody
	})
}
